#!/usr/bin/env python

# Attack Vector: Sunset Flight
# A rail shooter with volumetric pixel art.
# 2018-06-26 No Time To Play <https://notimetoplay.org/>

from __future__ import division

import os.path
import pygame

import inrelief
import game
import sprites

SCREENW = 1152
SCREENH = 720
CENTERX = SCREENW // 2
CENTERY = SCREENH // 2

MID_TOP = (CENTERX, 8)
BOTTOM_LEFT = (8, SCREENH - 8)
MID_BOTTOM = (CENTERX, SCREENH - 8)
BOTTOM_RIGHT = (SCREENW - 8, SCREENH - 8)

GAUGEW = SCREENW // 4
GAUGEH = SCREENH // 20
LGAUGEX = SCREENW // 8
RGAUGEX = CENTERX + SCREENW // 8

DARK = (20, 12, 28)
GRAPHITE = (78, 74, 78)
GRAY = (117, 113, 97)
MAUVE = (68, 36, 52)
BROWN = (133, 76, 48)
RED = (208, 70, 72)
ORANGE = (210, 125, 44)
YELLOW = (218, 212, 94)
LIME = (109, 170, 44)
GREEN = (52, 101, 36)
NAVY = (48, 52, 109)
BLUE = (89, 125, 206)
CYAN = (109, 194, 202)
STEEL = (133, 149, 161)
OCHRE = (210, 170, 153)
LIGHT = (222, 238, 214)

FIERY = (DARK, RED, ORANGE, YELLOW)
ELECTRIC = (DARK, BLUE, CYAN, LIGHT)
METAL = (DARK, GRAPHITE, GRAY, STEEL)
PLASMA = (DARK, GREEN, LIME, LIGHT)
SUNSET = (DARK, MAUVE, BROWN, RED)
EMBERS = (DARK, BROWN, RED, OCHRE)

GAMEDIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

pygame.mixer.pre_init(frequency = 44100)
pygame.init()

screen = pygame.display.set_mode((SCREENW, SCREENH))
clock = pygame.time.Clock()

def renderer(x, y, size, color):
	x += CENTERX
	y += CENTERY
	screen.fill(color, (x, y, size, size))

pygame.display.set_caption("Attack Vector: Sunset Flight")
pygame.event.set_allowed([pygame.QUIT, pygame.KEYDOWN, pygame.KEYUP,
	pygame.JOYHATMOTION, pygame.JOYBUTTONDOWN, pygame.JOYBUTTONUP])

if pygame.joystick.get_count() > 0:
	joystick = pygame.joystick.Joystick(0)
	joystick.init()
else:
	joystick = None

thin_font = os.path.join(
	GAMEDIR, "assets", "fonts", "kenvector_future_thin.ttf")
bold_font = os.path.join(
	GAMEDIR, "assets", "fonts", "kenvector_future.ttf")
hud_font = pygame.font.Font(thin_font, SCREENH // 15)
msg_font = pygame.font.Font(bold_font, SCREENH // 10)
big_font = pygame.font.Font(bold_font, SCREENH // 8)

menu_music = os.path.join(GAMEDIR, "assets", "audio", "menu-theme.ogg")
pause_music = os.path.join(GAMEDIR, "assets", "audio", "pause-theme.ogg")

sounds = ["synthetic_laser", "flak_gun_sound", "synthetic_bomb",
	"missile_alarm", "jetwash", "flak_hit", "bombexplosion",
	"tesla_tower", "radar_ready", "radar_fail"]
for i in range(len(sounds)):
	sounds[i] = pygame.mixer.Sound(
		os.path.join(GAMEDIR, "assets", "audio", sounds[i] + ".ogg"))
	sounds[i].set_volume(0.8)

S_START = 3
S_ENTER = 4
S_FLAK = 5
S_BOOM = 6
S_POWERUP = 7
S_MENUON = 8
S_MENUOFF = 9

weapon_icons = ["bullets-icon.png", "plasma-icon.png", "fireball-icon.png"]
for i in range(len(weapon_icons)):
	weapon_icons[i] = pygame.image.load(
		os.path.join(GAMEDIR, "assets", weapon_icons[i]))
	weapon_icons[i] = pygame.transform.scale(
		weapon_icons[i], (64, 64))

building1 = inrelief.Sprite(sprites.building, SUNSET)
building1.set_hotspot(8, 16, -2)
building2 = inrelief.Sprite(sprites.building2, EMBERS)
building2.set_hotspot(8, 16, -2)
building3 = inrelief.Sprite(sprites.building3, EMBERS)
building3.set_hotspot(8, 16, -2)

fireball = inrelief.Sprite(sprites.fireball, FIERY)
fireball.set_hotspot(4, 4, -2)
plasma = inrelief.Sprite(sprites.plasma, PLASMA)
plasma.set_hotspot(4, 4, -2)
bullet = inrelief.Sprite(sprites.bullet, ELECTRIC)
bullet.set_hotspot(4, 4, -2)

seeker = inrelief.Sprite(sprites.seeker, EMBERS)
seeker.set_hotspot(4, 4, -2)
powerup = inrelief.Sprite(sprites.powerup, PLASMA)
powerup.set_hotspot(8, 8, -2)

jet = inrelief.Sprite(sprites.jet, METAL)
jet.set_hotspot(8, 4, -2)
drone = inrelief.Sprite(sprites.drone, METAL)
drone.set_hotspot(8, 4, -2)
hovertank = inrelief.Sprite(sprites.hovertank, METAL)
hovertank.set_hotspot(8, 4, -2)

world = game.World()
world.load_buildings(building1, building2, building3)
world.load_weapons(bullet, plasma, fireball)
world.load_enemies(jet, drone, hovertank)
world.seeker_sprite = seeker
world.powerup_sprite = powerup

camera = inrelief.Camera(0, 0, 0, SCREENH)

def render_world(world, screen, camera):
	camera.x = world.player.x
	camera.y = world.player.y

	screen.fill(NAVY, (0, 0, SCREENW, CENTERY))
	screen.fill(DARK, (0, CENTERY, SCREENW, CENTERY))
	
	world.buildings.zsort()
	world.buildings.render_through(camera, renderer)
	
	# No need to sort first since they're just flat yellow discs.
	for i in world.explosions:
		if i["z"] > 0:
			camera.project(i["x"], i["y"], i["z"])
			pygame.draw.circle(
				screen, YELLOW,
				(int(camera.px) + CENTERX,
					int(camera.py) + CENTERY),
				int(i["age"] * camera.scale * 25))
	
	world.render_missiles(camera, renderer)

def render_hud(world, screen):
	#pygame.draw.line(screen, LIGHT, (0, CENTERY), (SCREENW - 1, CENTERY))
	#pygame.draw.line(screen, LIGHT, (CENTERX, 0), (CENTERX, SCREENH - 1))
	
	icon = weapon_icons[world.weapon]
	screen.blit(icon, icon.get_rect(midtop = MID_TOP))

	coords = (LGAUGEX, GAUGEH, GAUGEW, GAUGEH)
	pygame.draw.rect(screen, LIME, coords, 3)
	power = min(world.player.power / world.player.max_power, 1)
	coords = (LGAUGEX, GAUGEH, power * GAUGEW, GAUGEH)
	pygame.draw.rect(screen, LIME, coords)

	coords = (RGAUGEX, GAUGEH, GAUGEW, GAUGEH)
	pygame.draw.rect(screen, RED, coords, 3)
	heat = min(world.gun_heat / game.MAX_HEAT, 1)
	coords = (RGAUGEX, GAUGEH, heat * GAUGEW, GAUGEH)
	pygame.draw.rect(screen, RED, coords)
	
	try:
		fps = int(round(clock.get_fps()))
	except OverflowError as e:
		fps = 0
	fps_text = hud_font.render(str(fps), False, OCHRE)
	fps_pos = fps_text.get_rect(midbottom = MID_BOTTOM)
	screen.blit(fps_text, fps_pos)

	time_text = hud_font.render(str(int(world.time)), False, OCHRE)
	time_pos = time_text.get_rect(bottomleft = BOTTOM_LEFT)
	screen.blit(time_text, time_pos)
	
	score_text = hud_font.render(str(world.score), False, OCHRE)
	score_pos = score_text.get_rect(bottomright = BOTTOM_RIGHT)
	screen.blit(score_text, score_pos)
	
	if world.player.power <= 0:
		go_text = msg_font.render("GAME OVER", False, LIGHT)
		go_pos = go_text.get_rect(midbottom = (CENTERX, CENTERY))
		screen.blit(go_text, go_pos)
	if paused:
		p_text = msg_font.render("PAUSED", False, OCHRE)
		p_pos = p_text.get_rect(midtop = (CENTERX, CENTERY))
		screen.blit(p_text, p_pos)

running = True
paused = False
muted = False

title_timer = 0
score_timer = 10
gameover_timer = 3

high_scores = []
for i in range(5):
	high_scores.append({
		"game": "game #99",
		"time": 500 - i * 100,
		"points": 2500 - i * 500
	})

def add_high_score(game, timer, points):
	high_scores.append({
		"game": "game #%02d" % (game,),
		"time": timer,
		"points": points
	})
	high_scores.sort(
		key=lambda s: s["time"] * s["points"], reverse=True)
	while len(high_scores) > 5:
		high_scores.pop()

def handle_game_keydown(ev, world):
	if world.player.power <= 0:
		return
	elif ev.key == pygame.K_w or ev.key == pygame.K_UP:
		world.player.accel_y = -game.YMAX
	elif ev.key == pygame.K_a or ev.key == pygame.K_LEFT:
		world.player.accel_x = -game.XMAX
	elif ev.key == pygame.K_s or ev.key == pygame.K_DOWN:
		world.player.accel_y = game.YMAX
	elif ev.key == pygame.K_d or ev.key == pygame.K_RIGHT:
		world.player.accel_x = game.XMAX
	elif ev.key == pygame.K_SPACE:
		world.firing = True

def handle_game_keyup(ev, world):
	global running, paused, muted, loop_active_screen
	
	if ev.key == pygame.K_w or ev.key == pygame.K_UP:
		world.player.accel_y = 0
	elif ev.key == pygame.K_a or ev.key == pygame.K_LEFT:
		world.player.accel_x = 0
	elif ev.key == pygame.K_s or ev.key == pygame.K_DOWN:
		world.player.accel_y = 0
	elif ev.key == pygame.K_d or ev.key == pygame.K_RIGHT:
		world.player.accel_x = 0
	elif ev.key == pygame.K_SPACE:
		world.firing = False
	elif ev.key == pygame.K_1:
		world.weapon = 0
	elif ev.key == pygame.K_2:
		world.weapon = 1
	elif ev.key == pygame.K_3:
		world.weapon = 2
	elif ev.key == pygame.K_ESCAPE:
		sounds[S_MENUOFF].play()
		title_timer = 0
		loop_active_screen = loop_title_screen
		pygame.mixer.music.load(menu_music)
		if muted:
			pygame.mixer.music.set_volume(0.0)
		else:
			pygame.mixer.music.set_volume(0.7)
		pygame.mixer.music.play()
	elif ev.key == pygame.K_p:
		paused = not paused
		if paused:
			pygame.mixer.music.load(pause_music)
			if muted:
				pygame.mixer.music.set_volume(0.0)
			else:
				pygame.mixer.music.set_volume(0.7)
			pygame.mixer.music.play()
		else:
			pygame.mixer.music.stop()
	elif ev.key == pygame.K_m:
		muted = not muted
		if muted:
			for i in sounds:
				i.set_volume(0)
			pygame.mixer.music.set_volume(0.0)
		else:
			for i in sounds:
				i.set_volume(0.8)
			pygame.mixer.music.set_volume(0.7)

def return_to_title():
	global loop_active_screen, title_timer
	
	add_high_score(world.game_num, world.time, world.score)
	sounds[S_MENUOFF].play()
	title_timer = 0
	loop_active_screen = loop_title_screen
	pygame.mixer.music.load(menu_music)
	if muted:
		pygame.mixer.music.set_volume(0.0)
	else:
		pygame.mixer.music.set_volume(0.7)
	pygame.mixer.music.play()

def loop_game_screen(screen):
	global running, paused, gameover_timer

	# Simulations need the time in seconds, dammit!
	delta_t = clock.tick(60) * 0.001

	ev = pygame.event.poll()
	if ev.type == pygame.QUIT:
		running = False
	elif ev.type == pygame.KEYDOWN:
		handle_game_keydown(ev, world)
	elif ev.type == pygame.KEYUP:
		handle_game_keyup(ev, world)
	elif ev.type == pygame.JOYHATMOTION:
		if world.player.power <= 0:
			world.player.accel_x = 0
			world.player.accel_y = 0
		else:
			x, y = ev.value
			world.player.accel_x = x * game.XMAX
			world.player.accel_y = -y * game.YMAX
	elif ev.type == pygame.JOYBUTTONDOWN:
		if world.player.power <= 0:
			world.firing = False
		elif 0 <= ev.button <= 2:
			world.weapon = ev.button
			world.firing = True
	elif ev.type == pygame.JOYBUTTONUP:
		if 0 <= ev.button <= 2:
			world.firing = False
		elif ev.button == 3:
			paused = not paused
			if paused:
				pygame.mixer.music.load(pause_music)
				if muted:
					pygame.mixer.music.set_volume(0.0)
				else:
					pygame.mixer.music.set_volume(0.7)
				pygame.mixer.music.play()
			else:
				pygame.mixer.music.stop()
	
	if not paused:
		world.update(delta_t)
		if world.weapon_fired:
			sounds[world.weapon].play()
		if world.enemy_spawned:
			sounds[S_ENTER].play()
		if world.enemy_hit:
			sounds[S_FLAK].play()
		if world.enemy_destroyed or world.player_hit:
			sounds[S_BOOM].play()
		if world.powerup_touched:
			sounds[S_POWERUP].play()
		if world.player.power <= 0:
			gameover_timer -= delta_t
			if gameover_timer <= 0:
				return_to_title()

	render_world(world, screen, camera)
	render_hud(world, screen)
	
	pygame.display.flip()

def render_title(screen):
	text = hud_font.render("No Time To Play", False, OCHRE)
	pos = text.get_rect(midbottom = (CENTERX, CENTERY // 2))
	screen.blit(text, pos)
	text = hud_font.render("presents", False, OCHRE)
	pos = text.get_rect(midtop = (CENTERX, CENTERY // 2))
	screen.blit(text, pos)

	text = msg_font.render("Attack Vector", False, LIGHT)
	pos = text.get_rect(midbottom = (CENTERX, CENTERY))
	screen.blit(text, pos)
	text = big_font.render("SUNSET FLIGHT", False, LIGHT)
	pos = text.get_rect(midtop = (CENTERX, CENTERY))
	screen.blit(text, pos)

	if int(title_timer) % 2 == 0:
		text = hud_font.render("press fire", False, RED)
		pos = text.get_rect(
			center = (CENTERX, SCREENH - CENTERY // 2))
		screen.blit(text, pos)

def render_scores(screen):
	text = msg_font.render("High Scores", False, LIGHT)
	pos = text.get_rect(center = (CENTERX, CENTERY // 2))
	screen.blit(text, pos)
	
	for i in range(len(high_scores)):
		score = high_scores[i]
		line = "%s: %05d' %05d" % (
			score["game"], score["time"], score["points"])
		text = hud_font.render(line, False, OCHRE)
		pos = text.get_rect(
			midtop = (CENTERX, CENTERY + i * SCREENH // 15))
		screen.blit(text, pos)

def loop_title_screen(screen):
	global running, muted, loop_active_screen
	global title_timer, gameover_timer

	# Simulations need the time in seconds, dammit!
	delta_t = clock.tick(60) * 0.001

	ev = pygame.event.poll()
	if ev.type == pygame.QUIT:
		running = False
	elif ev.type == pygame.KEYUP:
		if ev.key == pygame.K_SPACE:
			world.start_game()
			gameover_timer = 3
			loop_active_screen = loop_game_screen
			pygame.mixer.music.stop()
			sounds[S_START].play(4)
		elif ev.key == pygame.K_m:
			muted = not muted
			if muted:
				for i in sounds:
					i.set_volume(0)
				pygame.mixer.music.set_volume(0.0)
			else:
				for i in sounds:
					i.set_volume(0.8)
				pygame.mixer.music.set_volume(0.7)
				sounds[S_MENUON].play()
		elif ev.key == pygame.K_ESCAPE:
			running = False
		elif ev.key == pygame.K_r:
			if world.game_num == 0:
				world.start_game()
			loop_active_screen = loop_game_screen
			pygame.mixer.music.stop()
	elif ev.type == pygame.JOYBUTTONUP:
		if 0 <= ev.button <= 2:
			world.start_game()
			gameover_timer = 3
			loop_active_screen = loop_game_screen
			pygame.mixer.music.stop()
			sounds[S_START].play(4)

	screen.fill(NAVY, (0, 0, SCREENW, CENTERY))
	screen.fill(DARK, (0, CENTERY, SCREENW, CENTERY))
	
	world.buildings.zsort()
	world.buildings.render_through(camera, renderer)

	title_timer += delta_t
	if title_timer >= 60:
		title_timer = 0
		pygame.mixer.music.play()

	if int(title_timer) % 20 < 10:
		render_title(screen)
	else:
		render_scores(screen)
	
	pygame.display.flip()

pygame.mixer.music.load(menu_music)
pygame.mixer.music.set_volume(0.7)
pygame.mixer.music.play()

loop_active_screen = loop_title_screen

while running:
	loop_active_screen(screen)
