# Attack Vector: Sunset Flight
# A rail shooter with volumetric pixel art.
# 2018-06-28 No Time To Play <https://notimetoplay.org/>

from __future__ import division

import math
import random
import inrelief

XMIN = -20
XMAX = 20
YMAX = 6
ZMAX = 300
MAX_HEAT = 60

class Vector3D:
	def __init__(self, x, y, z):
		self.x = x
		self.y = y
		self.z = z

	def __isub__(self, v):
		self.x -= v.x
		self.y -= v.y
		self.z -= v.z
		
		return self
	
	def magnitude(self):
		return math.sqrt(
			self.x ** 2 + self.y ** 2 + self.z ** 2)
	
	def normalized(self):
		m = self.magnitude()
		return Vector3D(self.x / m, self.y / m, self.z / m)

class Missile(inrelief.SpriteInstance):
	def __init__(self, sprite, x, y, z):
		inrelief.SpriteInstance.__init__(self, sprite, x, y, z)
		
		self.speed_x = 0
		self.speed_y = 0
		self.speed_z = 0
		
		self.accel_x = 0
		self.accel_y = 0
		
		self.friction = 0
		
		self.size_x = 0
		self.size_y = 0
		self.size_z = 0
		
		self.power = 0
		self.max_power = 0
		self.cooldown = 0

	def touches(self, m):
		if abs(self.x - m.x) > self.size_x + m.size_x:
			return False
		elif abs(self.y - m.y) > self.size_y + m.size_y:
			return False
		elif abs(self.z - m.z) > self.size_z + m.size_z:
			return False
		else:
			return True

	def aim_at(self, m):
		v = Vector3D(m.x, m.y, m.z)
		v -= Vector3D(self.x, self.y, self.z)
		return v.normalized()

	def update(self, delta_t):
		self.speed_x += self.accel_x * delta_t
		self.speed_y += self.accel_y * delta_t
		
		self.x += self.speed_x * delta_t
		self.y += self.speed_y * delta_t
		self.z += self.speed_z * delta_t
		
		self.speed_x -= self.speed_x * self.friction
		self.speed_y -= self.speed_y * self.friction
		
		if self.cooldown > 0:
			self.cooldown -= delta_t

class World:
	def __init__(self):
		self.buildings = inrelief.DisplayList()
		self.missiles = inrelief.DisplayList()
		
		self.weapon_data = [None, None, None]
		self.enemy_data = [None, None, None]
		self.seeker_sprite = None
		self.powerup_sprite = None
		
		self.shots = []
		self.enemies = []
		self.explosions = []
		
		self.player = Missile(None, 0, 3, 0)
		self.player.friction = 0.05
		self.player.size_x = 5
		self.player.size_y = 3
		self.player.size_z = 5
		
		self.game_num = 0
	
	def start_game(self):
		del self.shots[:]
		del self.enemies[:]
		del self.explosions[:]
		
		self.player.power = 60
		self.player.max_power = 60
		
		self.weapon = 0
		self.gun_heat = 0
		self.weapon_cooldown = 0
		self.enemy_cooldown = 2
		self.enemy_counter = 0
		
		self.time = 0
		self.score = 0
		self.game_num += 1
		
		self.firing = False
		
		# Flags for the sound system.
		self.weapon_fired = False
		self.enemy_spawned = False
		self.enemy_hit = False
		self.enemy_destroyed = False
		#self.enemy_passed = False
		self.seeker_spawned = False
		self.powerup_touched = False
		self.player_hit = False
		#self.player_destroyed = False
	
	def load_buildings(self, building1, building2, building3):
		sprites = [building1, building2, building3, building1,
			building1, building3, building2, building1,
			building1, building1]
		for i in range(ZMAX, 0, -60):
			self.buildings.append(sprites.pop(), -30, 8, i)
			self.buildings.append(sprites.pop(), 30, 8, i)
	
	def load_weapons(self, bullets, plasma, fireball):
		self.weapon_data[0] = {
			"sprite": bullets,
			"speed": 200,
			"size": 1,
			"cooldown": 0.15,
			"power": 2,
			"level": 3
		}
		self.weapon_data[1] = {
			"sprite": plasma,
			"speed": 150,
			"size": 2,
			"cooldown": 0.3,
			"power": 5,
			"level": 2
		}
		self.weapon_data[2] = {
			"sprite": fireball,
			"speed": 100,
			"size": 3,
			"cooldown": 0.6,
			"power": 8,
			"level": 1
		}
	
	def load_enemies(self, jet, drone, hovertank):
		self.enemy_data[0] = {
			"sprite": jet,
			"min_y": -6,
			"max_y": 0,
			"speed_z": -20,
			"power": 15
		}
		self.enemy_data[1] = {
			"sprite": drone,
			"min_y": -3,
			"max_y": 3,
			"speed_z": -15,
			"power": 25
		}
		self.enemy_data[2] = {
			"sprite": hovertank,
			"min_y": 0,
			"max_y": 6,
			"speed_z": -10,
			"power": 40
		}
	
	def spawn_shots(self):
		weapon = self.weapon_data[self.weapon]
		if weapon["level"] == 1:
			self.shots.append(self.make_shot(weapon, 0, 4))
		elif weapon["level"] == 2:
			self.shots.append(self.make_shot(weapon, -6, 4))
			self.shots.append(self.make_shot(weapon, 6, 4))
		elif weapon["level"] == 3:
			self.shots.append(self.make_shot(weapon, -6, 4))
			self.shots.append(self.make_shot(weapon, 0, -4))
			self.shots.append(self.make_shot(weapon, 6, 4))
		self.weapon_cooldown = weapon["cooldown"]
		self.gun_heat += weapon["power"] * weapon["level"]
		self.weapon_fired = True
	
	def make_shot(self, weapon, x, y):
		shot = Missile(
			weapon["sprite"],
			self.player.x + x,
			self.player.y + y,
			3)

		shot.speed_z = weapon["speed"]
		shot.power = weapon["power"]

		shot.size_x = weapon["size"]
		shot.size_y = weapon["size"]
		shot.size_z = weapon["size"]

		return shot
	
	def spawn_enemy(self):
		if self.enemy_counter % 3 == 0:
			e = self.enemy_data[0]
		elif self.enemy_counter % 2 == 0:
			e = self.enemy_data[1]
		else:
			e = self.enemy_data[2]
			
		enemy = Missile(
			e["sprite"],
			random.randint(XMIN, XMAX),
			random.randint(e["min_y"], e["max_y"]),
			ZMAX)

		enemy.speed_z = e["speed_z"]
		enemy.power = e["power"]
		enemy.max_power = e["power"]
		
		enemy.cooldown = enemy.max_power / 5

		enemy.size_x = 5
		enemy.size_y = 3
		enemy.size_z = 5

		self.enemies.append(enemy)
		self.enemy_spawned = True
	
	def spawn_seeker(self, e):
		seeker = Missile(self.seeker_sprite, e.x, e.y, e.z)

		v = e.aim_at(self.player)
		seeker.speed_x = v.x * 50
		seeker.speed_y = v.y * 50
		seeker.speed_z = v.z * 50
		
		seeker.power = 5
		seeker.max_power = 5
		
		seeker.size_x = 2
		seeker.size_y = 2
		seeker.size_z = 2
		
		self.enemies.append(seeker)
		self.seeker_spawned = True
	
	def spawn_powerup(self, x, y, z):
		powerup = Missile(self.powerup_sprite, x, y, z)
		powerup.speed_z = -50
		powerup.size_x = 6
		powerup.size_y = 6
		powerup.size_z = 6
		powerup.power = 100 # But leave the max power at zero.
		self.enemies.append(powerup)
	
	def spawn_explosion(self, x, y, z, max_age):
		self.explosions.append({
			"x": x,
			"y": y,
			"z": z,
			"age": 0,
			"max_age": max_age
		})
	
	def handle_enemy_shot(self, e, s):
		e.power -= s.power
		s.power = 0
		self.spawn_explosion(s.x, s.y, s.z, 0.15)
		if e.power <= 0:
			self.score += e.max_power
			self.spawn_explosion(e.x, e.y, e.z, 0.4)
			if random.random() < 0.1:
				self.spawn_powerup(e.x, e.y, e.z)
			self.enemy_destroyed = True
		else:
			self.enemy_hit = True
	
	def handle_enemy_crash(self, e):
		self.player.power -= e.max_power
		if self.player.power < 0:
			self.player.power = 0
		e.power = 0
		self.spawn_explosion(e.x, e.y, e.z, 0.7)
		self.enemy_destroyed = True
	
	def handle_powerup(self, p):
		self.gun_heat *= 0.5
		diff = self.player.max_power / 10
		self.player.power += diff
		if self.player.power > self.player.max_power:
			self.player.power = self.player.max_power
		p.power = 0
		self.powerup_touched = True
	
	def update(self, delta_t):
		self.weapon_fired = False
		self.enemy_spawned = False
		self.enemy_hit = False
		self.enemy_destroyed = False
		#self.enemy_passed = False
		self.seeker_spawned = False
		self.powerup_touched = False
		self.player_hit = False
		#self.player_destroyed = False
		
		p = self.player

		if abs(p.x) > XMAX:
			p.accel_x = -p.x * 0.5
		if abs(p.y) > YMAX:
			p.accel_y = -p.y * 0.5
		p.update(delta_t)
		
		for i in self.shots:
			i.z += i.speed_z * delta_t
		if len(self.shots) > 0:
			if self.shots[0].z > ZMAX:
				del self.shots[0]
		if self.firing:
			if self.weapon_cooldown <= 0:
				if self.gun_heat < MAX_HEAT:
					self.spawn_shots()
		if self.weapon_cooldown > 0:
			self.weapon_cooldown -= delta_t
		if self.gun_heat > delta_t * 10:
			self.gun_heat -= delta_t * 10
		
		for i in self.enemies:
			#i.z += i.speed_z * delta_t
			i.update(delta_t)
		self.enemies = [i for i in self.enemies if i.z > 0]
		
		for j in self.enemies:
			for i in self.shots:
				if j.max_power > 0 and i.touches(j):
					self.handle_enemy_shot(j, i)
			if j.touches(self.player):
				if j.max_power > 0:
					self.handle_enemy_crash(j)
				else: # Must be a powerup.
					self.handle_powerup(j)
			elif j.cooldown <= 0 and j.max_power > 5:
				self.spawn_seeker(j)
				j.cooldown = j.max_power / 5
		if p.power < p.max_power:
			if p.power > 0: # If dead, stay dead.
				p.power += delta_t * 0.2
		
		self.shots = [i for i in self.shots if i.power > 0]
		self.enemies = [i for i in self.enemies if i.power > 0]

		if self.enemy_cooldown > 0:
			if len(self.enemies) < 5:
				self.enemy_cooldown -= delta_t
		else:
			if self.enemy_counter % 20 < 15:
				self.spawn_enemy()
			self.enemy_cooldown = 2
			self.enemy_counter += 1
		
		for i in self.explosions:
			i["age"] += delta_t
		self.explosions = [i for i in self.explosions
			if i["age"] <= i["max_age"]]
		
		for i in self.buildings.sprites:
			i.z -= 30 * delta_t
			if i.z <= 0:
				i.z = ZMAX
		
		self.time += delta_t

	def render_missiles(self, camera, renderer):
		m = self.missiles
		del m.sprites[:]
		m.sprites += self.shots
		m.sprites += self.enemies
		m.zsort()
		m.render_through(camera, renderer)
